#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")] // hide console window on Windows in release

mod new_roll;
mod roll_view;

use eframe::egui;
use egui_datepicker::{Local, NaiveDate};
use new_roll::NewRollWindow;
use roll_view::RollViewWindow;
use serde::{Deserialize, Serialize};

fn main() -> Result<(), eframe::Error> {
    env_logger::init(); // Log to stderr (if you run with `RUST_LOG=debug`).
    let options = eframe::NativeOptions {
        initial_window_size: Some(egui::vec2(320.0, 240.0)),
        ..Default::default()
    };
    eframe::run_native(
        "Film Manager",
        options,
        Box::new(|cc| Box::new(MyApp::new(cc))),
    )
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Exposure {
    num: u8,
    #[serde(default)]
    printed: bool,
    #[serde(default)]
    damaged: bool,
    #[serde(default)]
    paper: Option<String>,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Roll {
    id: String,
    name: String,
    #[serde(default)]
    desc: String,
    date: NaiveDate,
    exposures: Vec<Exposure>,
}

impl Roll {
    fn new(id: String, name: String, desc: String, date: NaiveDate, num_exposures: u8) -> Self {
        let mut exposures = Vec::new();
        for num in 1..=num_exposures {
            exposures.push(Exposure {
                num,
                printed: false,
                damaged: false,
                paper: None,
            })
        }
        Self {
            id,
            name,
            desc,
            date,
            exposures,
        }
    }

    pub fn short_str(&self) -> String {
        format!(
            "{}{} on {}, {} exp",
            self.id,
            if self.name.is_empty() {
                String::new()
            } else {
                format!(" - {}", self.name)
            },
            self.date.format("%m/%d/%Y"),
            self.exposures.len()
        )
    }
}

pub struct AppState {
    pub rolls: Vec<Roll>,
}

struct MyApp {
    state: AppState,
    roll_views: Vec<RollViewWindow>,
    new_roll_window: Option<NewRollWindow>,
}

impl MyApp {
    fn new(cc: &eframe::CreationContext<'_>) -> Self {
        let rolls = eframe::get_value(cc.storage.unwrap(), "rolls").unwrap_or_else(|| {
            vec![Roll::new(
                "00001".into(),
                "Test".into(),
                "Testing".into(),
                Local::now().date_naive(),
                24,
            )]
        });
        Self {
            state: AppState { rolls },
            roll_views: Vec::new(),
            new_roll_window: None,
        }
    }
}

impl eframe::App for MyApp {
    fn save(&mut self, storage: &mut dyn eframe::Storage) {
        eframe::set_value(storage, "rolls", &self.state.rolls);
    }

    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        egui::CentralPanel::default().show(ctx, |ui| {
            ui.heading("Film Manager");
            ui.label("Rolls:");
            for (i, roll) in self.state.rolls.iter().enumerate() {
                if ui.button(roll.short_str()).clicked()
                    && self.roll_views.iter().find(|v| v.roll() == i).is_none()
                {
                    self.roll_views.push(RollViewWindow::new(i));
                }
            }
            if ui.button("Add").clicked() && self.new_roll_window.is_none() {
                self.new_roll_window = Some(NewRollWindow::default());
            }
            self.roll_views.retain_mut(|win| {
                !win.draw(ctx, &mut self.state)
            });
            if let Some(new_roll_window) = self.new_roll_window.as_mut() {
                if new_roll_window.draw(ctx, &mut self.state) {
                    self.new_roll_window = None;
                }
            }
        });
    }
}
